/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *	Mark Anderson, Argonne National Laboratory:
 *		U.S. DOE, University of Chicago
 */


#define ALLOCATE_STORAGE

#include <unistd.h>
#include <stdio.h>
/* If not MS Visual C++ or MS Visual C++ is 2010 or later  */
#if !defined(_MSC_VER) || _MSC_VER >= 1600
#include <stdint.h>
#endif

#include "dp.h"
#include <Xm/RepType.h>

#include "trashCan.bm"			/* pixmap for scissors */

#define N_MAX_MENU_ELES			8

#define N_MAIN_MENU_ELES		4

#define N_FILE_MENU_ELES		7
#define FILE_BTN_POSN			0
#define FILE_NEW_BTN			0
#define FILE_OPEN_BTN			1
#define FILE_SAVE_AS_BTN		2
#define FILE_PRINT_DATA_BTN		3
#define FILE_EXIT_BTN			4

#define N_EDIT_MENU_ELES		3
#define EDIT_BTN_POSN			1
#define EDIT_CUT_BTN			0
#define EDIT_CLEAR_BTN			1

#define N_STATISTICS_MENU_ELES		2
#define STATISTICS_BTN_POSN		2
#define STATISTICS_CLEAR_BTN		0
#define STATISTICS_CLEAR_ALL_BTN	1

#define N_HELP_MENU_ELES		7
#define HELP_BTN_POSN			3
#define HELP_ON_CONTEXT_BTN		0
#define HELP_ON_WINDOW_BTN		1
#define HELP_ON_KEYS_BTN		2
#define HELP_INDEX_BTN			3
#define HELP_ON_HELP_BTN		4
#define HELP_TUTORIAL_BTN		5
#define HELP_ON_VERSION_BTN		6



/* function to do all DND drop site registration, handling */
extern void addDropSite(Widget w);
extern void addTrashDropSite(Widget w);

extern void getCompositeHeadTail(DpStruct *, DpStruct **, DpStruct **);

extern void insertElementAfter(char *, DpStruct *);
extern void freeCAresources(void);
extern void removeElements(DpStruct *, Boolean);
extern void dpInitializeCA(void);

extern Widget createAndPopupProductDescriptionShell(
  XtAppContext appContext, Widget topLevelShell,
  char *name, XmFontList nameFontList,
  Pixmap namePixmap,
  char *description, XmFontList descriptionFontList,
  char *versionInfo, char *developedAt, XmFontList otherFontList,
  int background, int foreground, int seconds);

static void createMain(Widget);
extern void createDisplayPage(char *);
extern void clearStatistics(void);
extern void clearAllStatistics(void);
extern Widget createPrintDialog(Widget);

int main(int argc, char *argv[]);
static void createMain(Widget mainShell);
static void saveToFile(char *filename);
int getNumberSelected();
unsigned long getPixelFromString(Display *display, int screen,
  char *colorString);
void selectRow(Widget w, XtPointer client_data, XEvent *event);
void unselectRows(void);
void unselectRow(DpStruct *head, DpStruct *tail);
static void interactionTextFCallback(Widget w, XtPointer client_data,
  XtPointer call_data);
static void scrollCB(Widget w, XtPointer client_data, XtPointer call_data);
static void fileMenuSimpleCallback(Widget w, XtPointer client_data,
  XtPointer call_data);
static void editMenuSimpleCallback(Widget w, XtPointer client_data,
  XtPointer call_data);
static void statisticsMenuSimpleCallback(Widget w, XtPointer client_data,
  XtPointer call_data);
static void helpMenuSimpleCallback(Widget w, XtPointer client_data,
  XtPointer call_data);
static void helpDialogCallback(Widget w, XtPointer client_data,
  XtPointer call_data);
static void fileMenuDialogCallback(Widget w, XtPointer client_data,
  XtPointer call_data);



/*
 * globals for use by any routines in this file
 */
static Widget openFSD = NULL;
static Widget productDescriptionShell = NULL;
static XmString openXmString = NULL;
static XmString saveAsXmString = NULL;

static String fallbackResources[] = {

  "Dp*initialResourcesPersistent: False",
  "Dp*keyboardFocusPolicy: XmPOINTER",
  "Dp*foreground: black",
  "Dp*background: #b0c3ca",
  "Dp*selectThickness: 1",
  "Dp*shadowThickness: 2",
  "Dp*fontList: 9x15",
  "Dp*tearOffModel: XmTEAR_OFF_ENABLED",

  "Dp.mainMW*mainBB1*DA*compositeL*fontList: 8x13bold",
  "Dp.mainMW*mainBB1*DA*channelL.fontList: 8x13",
  "Dp.mainMW*mainBB1*fontList: 8x13",
  "Dp.mainMW*legendBB*fontList: 8x13",

/***
 *** main window
 ***/
  "Dp.mainMW.mainMB*fontList: 9x15",
  "Dp.mainMW.width: 600",
  "Dp.mainMW.height: 400",
  "Dp.mainMW*legendBB.width: 600",
  "Dp.mainMW*legendBB.height: 32",
  "Dp.mainMW*mainBB.width: 600",
  "Dp.mainMW*mainBB.height: 350",

  "Dp.mainMW.mainBB.background: white",
  "Dp.mainMW.mainBB.foreground: white",
  "Dp.mainMW.mainBB.highlightColor: grey90",
  "Dp.mainMW.legendBB*background: grey95",
  "Dp.mainMW.mainBB*background: grey95",
  "Dp.mainMW.mainBB*selectOnEnter: False",
  "Dp.mainMW.mainBB*highlightColor: grey65",
  "Dp.mainMW*mainBB1*foreground: black",
  "Dp.mainMW*mainBB1*background: white",
  "Dp.mainMW*mainBB1*setPointTF.topShadowColor: grey70",
  "Dp.mainMW*mainBB1*setPointTF.bottomShadowColor: grey70",
  "Dp.mainMW*legendBB.shadowThickness: 0",

  "Dp.mainMW*trashCan.background: #90a3aa",
  "Dp.mainMW*trashCan.shadowType: XmSHADOW_ETCHED_OUT",

  "Dp*XmBulletinBoard.marginWidth: 0",
  "Dp*XmBulletinBoard.marginHeight: 0",
  "Dp*XmDrawingArea.marginWidth: 0",
  "Dp*XmDrawingArea.marginHeight: 0",
  "Dp*XmMainWindow.showSeparator: False",
  "Dp.mainMW*DA.setPointTF.columns: 10",
  "Dp.mainMW*DA.XmTextField.columns: 12",	/* dp.h: NOMINAL_NCOLUMNS=12 */
  "Dp.mainMW*DA.XmTextField.marginHeight: 0",
  "Dp.mainMW*DA.XmTextField.marginWidth: 1",
  "Dp.mainMW*DA.simpleL.marginHeight: 2",
  "Dp.mainMW*DA.simpleL.labelString: ",
  "Dp.mainMW*DA.simpleL.alignment: XmALIGNMENT_BEGINNING",

  "Dp.mainMW*interactionTextF.shadowThickness: 2",

  "Dp*XmTextField.translations: #override \
    None<Key>osfDelete:     delete-previous-character() \n\
    None<Key>osfBackSpace:  delete-next-character()",

  "Dp*XmTextField.verifyBell: False",

  (String)NULL
};


/*********
 ********* MISC FUNCTIONS
 *********/

/*
 * function to save current device/display information to specified DP file
 */

static void saveToFile(char *filename)
{
  FILE *stream;
  DpStruct *dpStruct;
  Arg args[2];
  char *buffer;
  XmString xmString;
  int n;

  stream = fopen(filename,"w+");

  if (stream != (FILE *)NULL) {

    dpStruct = listHead;
    while (dpStruct != (DpStruct *)NULL) {
/* output composite names, atomics not members of composites, and PVs only */
      if (dpStruct->entryType == CompositeDevice ||
	dpStruct->entryType == ProcessVariable ||
	(dpStruct->entryType == AtomicDevice && !dpStruct->memberOfComposite)) {
        fprintf(stream,"%s\n",dpStruct->channel);
      }

      dpStruct = dpStruct->next;
    }

    fclose(stream);

  } else {
    fprintf(stderr,"\nsaveToFile: cannot save to file: %s",filename);
/* leave room for warning string below (Cannot save...) */
    buffer = (char *)malloc((strlen(filename)+40)*sizeof(char));
    sprintf(buffer,"Cannot save to file:  %s  ",filename);
    xmString = XmStringCreateSimple(buffer);
    n = 0;
    XtSetArg(args[n],XmNselectionLabelString,xmString); n++;
    XtSetValues(openFSD,args,n);
    XtManageChild(openFSD);
    XmStringFree(xmString);
    free((char *)buffer);
  }

}




/*
 * function to return number of currently selected rows
 */
int getNumberSelected()
{
  int numSelected = 0;

  if (currentlySelectedDpStructs != (DpStruct **)NULL) {
    while (currentlySelectedDpStructs[numSelected] != (DpStruct *)NULL) {
	numSelected++;
    }
  }
  return (numSelected);
}


/*
 * function to return a pixel value from a specified color string
 */
unsigned long getPixelFromString(Display *display, int screen,
  char *colorString)
{
  XColor color, ignore;

  if (!XAllocNamedColor(display,DefaultColormap(display,screen),
		     colorString,&color,&ignore)) {
      fprintf(stderr, "\ngetPixelFromString:  couldn't allocate color %s",
					     colorString);
      return(WhitePixel(display, screen));
   } else {
      return(color.pixel);
   }
}




/************************************************
 **************** EventHandlers *****************
 ************************************************/


void selectRow(Widget w, XtPointer client_data, XEvent *event)
{
  Pixel highlightColor;
  Arg args[4];
  DpStruct *dpStruct, *head, *tail, *ptr;
  Widget DA;
  WidgetList children;
  Cardinal numChildren;
  int i, numSelected;
  XButtonEvent *xbutton;
  XtPointer userData;
  Boolean alreadyInList;

  xbutton = (XButtonEvent *)event;

/* valid selection (MB1) */
  if (xbutton->button == Button1) {

    if (XtClass(w) == xmDrawingAreaWidgetClass) DA = w;
    else DA = XtParent(w);
    XtVaGetValues(DA,XmNuserData,&userData,NULL);
    dpStruct = (DpStruct *)userData;

/* get head and tail of list if composite, else atomic or pv */
    getCompositeHeadTail(dpStruct,&head,&tail);

  /* not shifted, therefore reset selected list */
    if (!(xbutton->state & ShiftMask) ) {

      if (isCurrentlySelected(head)) {
      /* unselect (toggle selection) of selected element */
	unselectRow(head,tail);
	alreadyInList = True;
      } else {
      /* unselect any currently selected rows */
	unselectRows();
	currentlySelectedDpStructs = (DpStruct **)malloc(
			2*sizeof(DpStruct *));
	currentlySelectedDpStructs[0] = head;
	currentlySelectedDpStructs[1] = (DpStruct *)NULL;
	alreadyInList = False;
      }

    } else {

      numSelected = getNumberSelected();


    /* make sure this row isn't already selected */
      alreadyInList = False;
      for (i = 0; i < numSelected; i++) {
	if (head == currentlySelectedDpStructs[i])
	   alreadyInList = True;
      }
      if (!alreadyInList) {
    /* augment any currently selected rows */
	if (numSelected == 0) {
          currentlySelectedDpStructs = (DpStruct **)malloc(
		(2)*sizeof(DpStruct *));
	} else {
          currentlySelectedDpStructs = (DpStruct **)realloc(
		currentlySelectedDpStructs,(numSelected+2)*sizeof(DpStruct *));
	}
        currentlySelectedDpStructs[numSelected] = head;
        currentlySelectedDpStructs[numSelected+1] = (DpStruct *)NULL;
      }
    }

    if (!alreadyInList) {
      if (head->entryType == CompositeDevice) {
        XtSetArg(args[0],XmNhighlightColor,&highlightColor);
        XtGetValues(mainBB,args,1);
	ptr = head;
	/* loop over all entries in composite */
	while (ptr != tail->next) {
          XtVaGetValues(ptr->DA,XmNchildren,&children,
		XmNnumChildren,&numChildren,NULL);
          XtSetArg(args[0],XmNbackground,highlightColor);
          XtSetValues(ptr->DA,args,1);
          for (i = 0; i < numChildren; i++) {
	    XtSetValues(children[i],args,1);
          }
	  ptr = ptr->next;
	}
      } else {
	/* not composite, do the selected element */
        XtSetArg(args[0],XmNhighlightColor,&highlightColor);
        XtGetValues(mainBB,args,1);
        XtVaGetValues(head->DA,XmNchildren,&children,
		XmNnumChildren,&numChildren,NULL);
        XtSetArg(args[0],XmNbackground,highlightColor);
        XtSetValues(head->DA,args,1);
        for (i = 0; i < numChildren; i++) {
	  XtSetValues(children[i],args,1);
        }
      }
    } /* if (!alreadyInList) */


  } /* if (xbutton->button == Button1) */
}



/* simple unselect function (not event handler) */
void unselectRows(void)
{
  Pixel background;
  Arg args[4];
  DpStruct *head, *tail, *ptr;
  WidgetList children;
  Cardinal numChildren;
  int i, j;

  if (currentlySelectedDpStructs != (DpStruct **)NULL) {
    XtSetArg(args[0],XmNbackground,&background);
    XtGetValues(mainBB,args,1);
    j = 0;
    while (currentlySelectedDpStructs[j] != (DpStruct *)NULL) {
      getCompositeHeadTail(currentlySelectedDpStructs[j],&head,&tail);
      ptr = head;
      while (ptr != tail->next) {
        XtVaGetValues(ptr->DA,XmNchildren,&children,
		XmNnumChildren,&numChildren,NULL);
        XtSetArg(args[0],XmNbackground,background);
        XtSetValues(ptr->DA,args,1);
        for (i = 0; i < numChildren; i++) {
          XtSetValues(children[i],args,1);
        }
	ptr = ptr->next;
      } 
      j++;
    }
    free((char *)currentlySelectedDpStructs);
    currentlySelectedDpStructs = (DpStruct **)NULL;
  }
}


/* simple unselect function to remove a row from selected list */
void unselectRow(DpStruct *head, DpStruct *tail)
{
  Pixel background;
  Arg args[4];
  DpStruct *ptr;
  WidgetList children;
  Cardinal numChildren;
  int i, j, numSelected;

  if (currentlySelectedDpStructs != (DpStruct **)NULL) {
    XtSetArg(args[0],XmNbackground,&background);
    XtGetValues(mainBB,args,1);
    ptr = head;
    while (ptr != tail->next) {
        XtVaGetValues(ptr->DA,XmNchildren,&children,
		XmNnumChildren,&numChildren,NULL);
        XtSetArg(args[0],XmNbackground,background);
        XtSetValues(ptr->DA,args,1);
        for (i = 0; i < numChildren; i++) {
          XtSetValues(children[i],args,1);
        }
	ptr = ptr->next;
    } 
  /*  now clear the entry from the list and shuffle everybody down */
    numSelected = getNumberSelected();
    i = 0;
    while (head != currentlySelectedDpStructs[i]) i++;
    for (j = i; j < numSelected; j++) {
       currentlySelectedDpStructs[j] = currentlySelectedDpStructs[j+1];
    }
    if (getNumberSelected()==0) {
      free((char *)currentlySelectedDpStructs);
      currentlySelectedDpStructs = (DpStruct **)NULL;
    }
  }
}





/********************************************
 **************** Callbacks *****************
 ********************************************/

/*
 * interactionTextF text field activate callback
 *   insert any elements resolved from string after
 *	currentlySelectedDpStructs
 */
static void interactionTextFCallback(Widget w, XtPointer client_data,
  XtPointer call_data)
{
  DpStruct *dpStruct;
  char *string;
  int i, num;
  Boolean foundIt;

  dpStruct = listTail;
  string = XmTextFieldGetString(w);
  if (currentlySelectedDpStructs != (DpStruct **)NULL) {
    num = getNumberSelected();
  /* really want to append after last (bottom of list) selected element */
    if (dpStruct != NULL) {
      foundIt = False;
      while (dpStruct != listHead && !foundIt) {
	for (i = num-1; i >= 0; i--) {
	  if (dpStruct == currentlySelectedDpStructs[i]) {
	    foundIt = True;
	    break;
	  }
	}
	if (!foundIt) dpStruct = dpStruct->prev;
      }
    }
  }
  insertElementAfter(string,dpStruct);
  XtFree(string);
  XmTextFieldSetString(w,"");
}


static void scrollCB(Widget w, XtPointer client_data, XtPointer call_data)
{
    XmScrollBarCallbackStruct *cbs=(XmScrollBarCallbackStruct *)call_data;
  /* important note:  XtMoveWidget can move widgets to negative positions,
   *	XtSetValues(..., XmNx/XmNy,...) cannot!
   */

  if (device == NULL) return;

  XtMoveWidget(device,(Position)(deviceX - cbs->value),2);
  XtMoveWidget(setPoint,(Position)(setPointX - cbs->value),2);
  XtMoveWidget(rdbk,(Position)(rdbkX - cbs->value),2);
  XtMoveWidget(average,(Position)(averageX - cbs->value),2);
  XtMoveWidget(rms,(Position)(rmsX - cbs->value),2);
  XtMoveWidget(diff5,(Position)(diff5X - cbs->value),2);
  XtMoveWidget(diff,(Position)(diffX - cbs->value),2);
  XtMoveWidget(units,(Position)(unitsX - cbs->value),2);
  XtMoveWidget(operRange,(Position)(operRangeX - cbs->value),2);
}


static void fileMenuSimpleCallback(Widget w, XtPointer client_data,
  XtPointer call_data)
{
    XmString dirMask;
    Arg args[4];
    int n;
    intptr_t buttonNumber=(intptr_t)client_data;


    switch(buttonNumber) {

	case FILE_NEW_BTN:		/* just like Edit->Clear */
		freeCAresources();
        /* if already existing data, delete it and continue */
                if (mainBB1 != NULL) {
                   XtDestroyWidget(mainBB1);
                   mainBB1 = (Widget)NULL;
                }
		break;

	case FILE_OPEN_BTN:
	case FILE_SAVE_AS_BTN:
		/* use userData as key to whether doing Open or Save */
		n = 0;
		XtSetArg(args[n],XmNuserData,(XtPointer)buttonNumber); n++;
		if (buttonNumber == FILE_OPEN_BTN) {
		  XtSetArg(args[n],XmNdialogTitle,openXmString); n++;
		  XtSetArg(args[n],XmNselectionLabelString,openXmString); n++;
		} else if (buttonNumber == FILE_SAVE_AS_BTN) {
		  XtSetArg(args[n],XmNdialogTitle,saveAsXmString); n++;
		  XtSetArg(args[n],XmNselectionLabelString,saveAsXmString); n++;
		}
		XtSetValues(openFSD,args,n);
		XtVaGetValues(openFSD,XmNdirMask,&dirMask,NULL);
		XmListDeselectAllItems(XmFileSelectionBoxGetChild(openFSD,
			XmDIALOG_LIST));
		XmFileSelectionDoSearch(openFSD,dirMask);
		XmStringFree(dirMask);
		XtManageChild(openFSD);
		break;

	case FILE_PRINT_DATA_BTN:
		if (printDialog == (Widget)NULL) {
		    printDialog = createPrintDialog(mainShell);
		}
		XtManageChild(printDialog);
		break;

	case FILE_EXIT_BTN:
		exit(0);
		break;
    }
}


static void editMenuSimpleCallback(Widget w, XtPointer client_data,
  XtPointer call_data)
{
    int buttonNumber=(intptr_t)client_data;

    switch(buttonNumber) {

	case EDIT_CLEAR_BTN:
		freeCAresources();
	/* if already existing data, delete it and continue */
		if (mainBB1 != NULL) {
		   XtDestroyWidget(mainBB1);
		   mainBB1 = (Widget)NULL;
		}
		break;

	case EDIT_CUT_BTN:
		if (getNumberSelected() > 0)
		    removeElements(currentlySelectedDpStructs[0],True);
		break;
    }
}



static void statisticsMenuSimpleCallback(Widget w, XtPointer client_data,
  XtPointer call_data)
{
    int buttonNumber=(intptr_t)client_data;

    switch(buttonNumber) {

	case STATISTICS_CLEAR_BTN:
		clearStatistics();
		break;
	case STATISTICS_CLEAR_ALL_BTN:
		clearAllStatistics();
		break;
    }
}




static void helpMenuSimpleCallback(Widget w, XtPointer client_data,
  XtPointer call_data)
{
    Widget widget;
    int buttonNumber=(intptr_t)client_data;
    XmAnyCallbackStruct *cbs=(XmAnyCallbackStruct *)call_data;

    switch(buttonNumber) {
    /* implement context sensitive help */
	case HELP_ON_CONTEXT_BTN:
		if ((widget = XmTrackingLocate(mainShell,helpCursor,False))) {
		  cbs->reason = XmCR_HELP;
		  XtCallCallbacks(widget,XmNhelpCallback,&call_data);
		}
		break;
	case HELP_ON_WINDOW_BTN:
	case HELP_ON_KEYS_BTN:
	case HELP_INDEX_BTN:
	case HELP_ON_HELP_BTN:
	case HELP_TUTORIAL_BTN:
		break;
	case HELP_ON_VERSION_BTN:
		XtPopup(productDescriptionShell,XtGrabNone);
		break;
    }
}



static void helpDialogCallback(Widget w, XtPointer client_data,
  XtPointer call_data)
{
    XmAnyCallbackStruct *cbs=(XmAnyCallbackStruct *)call_data;
    
    switch(cbs->reason){
    case XmCR_OK:
    case XmCR_CANCEL:
	XtPopdown(helpS);
	break;
    }
}



static void fileMenuDialogCallback(Widget w, XtPointer client_data,
  XtPointer call_data)
{
  char *filename;
  XtPointer userData;
  int n, len, btnNumber;
  Arg args[2];
  XmString xmString;
  XmFileSelectionBoxCallbackStruct *cbs=(XmFileSelectionBoxCallbackStruct *)call_data;

  switch(cbs->reason){
     case XmCR_CANCEL:
	XtUnmanageChild(w);
	break;

     case XmCR_OK:
/* get the filename string from the selection box */
	XmStringGetLtoR(cbs->value,XmSTRING_DEFAULT_CHARSET,&filename);
	len = strlen(filename);
/* check for proper suffix */
	if (filename[len-3] != '.' && filename[len-2] != 'd'
		&& filename[len-1] != 'p') {
      /* leave room for warning string below (Cannot save...) */
	  xmString = XmStringCreateSimple("Please use .dp suffix for filename");
	  n = 0;
	  XtSetArg(args[n],XmNselectionLabelString,xmString); n++;
	  XtSetValues(openFSD,args,n);
	  XtManageChild(openFSD);
	  XBell(XtDisplay(openFSD),100);
	  XmStringFree(xmString);
	} else {
	  XtUnmanageChild(w);
	  XtVaGetValues(w,XmNuserData,&userData,NULL);
	  btnNumber = (intptr_t)userData;
	  if (btnNumber == FILE_OPEN_BTN) {
	    createDisplayPage(filename);
	  } else if (btnNumber == FILE_SAVE_AS_BTN) {
	    XFlush(XtDisplay(w));
	    saveToFile(filename);
	  }
	}
	XtFree(filename);
	break;
  }
}

int main(int argc, char *argv[])
{
  int n;
  char versionString[60];
  XtGCMask gcValueMask;
  XGCValues gcValues;


/*
 * initialize a few globals
 */
 listHead = listTail = (DpStruct *)NULL;
 mainBB = mainBB1 = (Widget)NULL;
 printDialog = (Widget)NULL;
 /* legend label widgets */
 device = setPoint = average = rms = diff5 = diff = operRange = (Widget)NULL;
 numChannels = 0;
 intervalId = (XtIntervalId)NULL;
 currentlySelectedDpStructs = (DpStruct **)NULL;
 operRangeMaxColumns = NOMINAL_NCOLUMNS;

/*
 * to get around ca repeater task not running problem (with resultant
 *   orphaned windows), do ca_task_initialize() right away
 */
  SEVCHK(ca_task_initialize(),
                "\ndpInitializeCA: error in ca_task_initialize");

/*
 * initialize the Intrinsics..., create main shell
 */   
  n = 0;
  mainShell = XtAppInitialize(&appContext, CLASS, NULL, 0, &argc, argv,
		fallbackResources, NULL, n);

  display = XtDisplay(mainShell);
  if (display == NULL) {
        XtWarning("cannot open display");
        exit(-1);
  }

/* add necessary Motif resource converters */
/*
  XmRegisterConverters();
*/
  XmRepTypeInstallTearOffModelConverter();

/* add translations/actions for drag-and-drop */
  parsedTranslations = XtParseTranslationTable(dragTranslations);
  XtAppAddActions(appContext,dragActions,XtNumber(dragActions));


  WM_DELETE_WINDOW = XmInternAtom(display,"WM_DELETE_WINDOW",False);
  WM_TAKE_FOCUS =  XmInternAtom(display,"WM_TAKE_FOCUS",False);
  COMPOUND_TEXT =  XmInternAtom(display,"COMPOUND_TEXT",False);
  DP_DESTROY = XmInternAtom(display,"DP_DESTROY",False);
  screenNum = DefaultScreen(display);
  rootWindow = RootWindow(display,screenNum);


/* create standard font cursors */
  helpCursor = XCreateFontCursor(display,XC_question_arrow);
  watchCursor = XCreateFontCursor(display,XC_watch);

/* create initial XmStrings */
  openXmString = XmStringCreateSimple("Open... (*.dp)");
  saveAsXmString = XmStringCreateSimple("Save As... (*.dp)");

/*
 * create the main window area
 */
  createMain(mainShell);

/* create status GCs */
  gcValueMask = GCForeground;
  gcValues.foreground = getPixelFromString(display,screenNum,
	  BAD_STATUS_COLOR_STRING);
  badStatusGC = XtGetGC(mainShell,gcValueMask,&gcValues);
  gcValues.foreground = getPixelFromString(display,screenNum,
	GOOD_STATUS_COLOR_STRING);
  goodStatusGC = XtGetGC(mainShell,gcValueMask,&gcValues);
  gcValues.foreground = getPixelFromString(display,screenNum,
	BORDER_COLOR_STRING);
  borderGC = XtGetGC(mainShell,gcValueMask,&gcValues);



/*
 * and realize the toplevel shell widget
 */
  XtRealizeWidget(mainShell);

/*
 * setup a window to use for client message events 
 */
  clientMessageWindow = XtWindow(mainShell);


/*
 * initialize CA
 */
 dpInitializeCA();


/*
 * read the display page description file and generate the page
 */
  if (argc == 1 || argc == 2) {
    createDisplayPage((argc == 1 ? NULL : argv[1]) );
  } else {
    fprintf(stderr,"\ndp: you may supply no more than one display page description file as a command line argument\n");
    exit(-1);
  }


/*
 * create and popup the product description shell
 */
  sprintf(versionString,"%s  (%s)",DP_VERSION_STRING,EPICS_VERSION_STRING);
  productDescriptionShell = createAndPopupProductDescriptionShell(appContext,
	mainShell,
        "DP",
	NULL,	/* fontlist */
	(Pixmap)NULL,
        "Display Page",
	NULL, 	/* fontlist */
        versionString,
        "developed at Argonne National Laboratory, by Mark Anderson",
	NULL,	/* fontlist */
        -1, -1, 2);

/*
 * now go into event loop
 */

  XtAppMainLoop(appContext);

/*
  while (True) {
    XtAppNextEvent(appContext,&event);
    switch (event.type) {
      case ClientMessage:
		clientMessageEvent = (XClientMessageEvent *)(&event);
		if (clientMessageEvent->message_type == DP_DESTROY) {
		  widget = (Widget)(clientMessageEvent->data.l[0]);
		  if (widget != NULL) {
		     XtDestroyWidget(widget);
		  }
		} else {
		  XtDispatchEvent(&event);
		}
		break;
      default:
		XtDispatchEvent(&event);
		break;
    }
  }
*/



  return 0;
}

static void createMain(Widget mainShell)
{

  XmString buttons[N_MAX_MENU_ELES];
  KeySym keySyms[N_MAX_MENU_ELES];
  String accelerators[N_MAX_MENU_ELES];
  XmString acceleratorText[N_MAX_MENU_ELES];
  XmString label, cwdXmString;
  XmButtonType buttonType[N_MAX_MENU_ELES];
  Widget mainFilePDM, mainEditPDM, mainStatisticsPDM, mainHelpPDM;
  Widget menuHelpWidget;
  Widget interactionForm, subForm, interactionLabel, interactionTextF;
  XImage trashCanImage;
  Pixel fg, bg;
  char *cwd;

  int i, j, n;
  Arg args[12];

  n = 0;
/* create a main window child of the main shell */
  XtSetArg(args[n],XmNscrollBarDisplayPolicy,XmSTATIC); n++;
  XtSetArg(args[n],XmNscrollingPolicy,XmAUTOMATIC); n++;
  mainMW = XmCreateMainWindow(mainShell,"mainMW",args,n);

/* legend area */
  n = 0;
  legendBB = XmCreateBulletinBoard(mainMW,"legendBB",args,n);

/* work area */
  n = 0;
  XtSetArg(args[n],XmNshadowThickness,0); n++;
  XtSetArg(args[n],XmNresizePolicy,XmRESIZE_ANY); n++;
  mainBB = XmCreateBulletinBoard(mainMW,"mainBB",args,n);
/* add mainBB as valid drop site */
  addDropSite(mainBB);
/* add change event handler */



/* message window (drop site for deletes) */
  n = 0;
  XtSetArg(args[n],XmNhorizontalSpacing,4); n++;
  XtSetArg(args[n],XmNverticalSpacing,4); n++;
  XtSetArg(args[n],XmNshadowType,XmSHADOW_ETCHED_OUT); n++;
  interactionForm = XmCreateForm(mainMW,"interactionForm",args,n);

/* drop site for deletes */
  n = 0;
  XtSetArg(args[n],XmNrightAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNtopAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNbottomAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNwidth,trashCan_width); n++;
  XtSetArg(args[n],XmNheight,trashCan_height); n++;
  XtSetArg(args[n],XmNlabelType,XmPIXMAP); n++;
  trashCan = XmCreateDrawnButton(interactionForm,"trashCan",args,n);
  trashCanImage.width = trashCan_width;
  trashCanImage.height = trashCan_height;
/* MDA  forcing this a bit */
  trashCanImage.data = (char *)trashCan_bits;
  trashCanImage.xoffset = 0;
  trashCanImage.format = XYBitmap;
  trashCanImage.byte_order = MSBFirst;
  trashCanImage.bitmap_pad = 8;
  trashCanImage.bitmap_bit_order = LSBFirst;
  trashCanImage.bitmap_unit = 8;
  trashCanImage.depth = 1;
  trashCanImage.bytes_per_line = 5;
  trashCanImage.obdata = NULL;
  XmInstallImage(&trashCanImage,"trashCan");
  XtVaGetValues(trashCan,XmNforeground,&fg,XmNbackground,&bg,NULL);
  trashCanPixmap = XmGetPixmap(XtScreen(mainBB),"trashCan",fg,bg);
  n = 0;
  XtSetArg(args[n],XmNmarginHeight,0); n++;
  XtSetArg(args[n],XmNmarginTop,0); n++;
  XtSetArg(args[n],XmNmarginBottom,0); n++;
  XtSetArg(args[n],XmNalignment,XmALIGNMENT_BEGINNING); n++;
  XtSetArg(args[n],XmNlabelPixmap,trashCanPixmap); n++;
  XtSetValues(trashCan,args,n);
  addTrashDropSite(trashCan);
  XtManageChild(trashCan);

/* text entry area */
  n = 0;
  XtSetArg(args[n],XmNleftAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNrightAttachment,XmATTACH_WIDGET); n++;
  XtSetArg(args[n],XmNrightWidget,trashCan); n++;
  XtSetArg(args[n],XmNtopAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNbottomAttachment,XmATTACH_FORM); n++;
  subForm = XmCreateForm(interactionForm,"subform",args,n);
  XtManageChild(subForm);
  n = 0;
  label = XmStringCreateSimple("Append:");
  XtSetArg(args[n],XmNlabelType,XmSTRING); n++;
  XtSetArg(args[n],XmNlabelString,label); n++;
  XtSetArg(args[n],XmNleftAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNtopAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNbottomAttachment,XmATTACH_FORM); n++;
  interactionLabel = XmCreateLabel(subForm,"interactionLabel",args,n);
  XmStringFree(label);
  XtManageChild(interactionLabel);

  n = 0;
  XtSetArg(args[n],XmNeditable,True); n++;
  XtSetArg(args[n],XmNrightAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNleftAttachment,XmATTACH_WIDGET); n++;
  XtSetArg(args[n],XmNleftWidget,interactionLabel); n++;
  XtSetArg(args[n],XmNtopAttachment,XmATTACH_FORM); n++;
  XtSetArg(args[n],XmNbottomAttachment,XmATTACH_FORM); n++;
  interactionTextF = XmCreateTextField(subForm,"interactionTextF",
	args,n);
  XtAddCallback(interactionTextF,XmNactivateCallback,
	interactionTextFCallback,(XtPointer)NULL);
  XtManageChild(interactionTextF);

  

/*
 * create the menu bar
 */
  buttons[0] = XmStringCreateSimple("File");
  buttons[1] = XmStringCreateSimple("Edit");
  buttons[2] = XmStringCreateSimple("Statistics");
  buttons[3] = XmStringCreateSimple("Help");
  keySyms[0] = 'F';
  keySyms[1] = 'E';
  keySyms[2] = 'S';
  keySyms[3] = 'H';
  n = 0;
  XtSetArg(args[n],XmNbuttonCount,N_MAIN_MENU_ELES); n++;
  XtSetArg(args[n],XmNbuttons,buttons); n++;
  XtSetArg(args[n],XmNbuttonMnemonics,keySyms); n++;
  mainMB = XmCreateSimpleMenuBar(mainMW,"mainMB",args,n);

/* dig out the Help cascade button  and set the resource in the menu bar */
  menuHelpWidget = XtNameToWidget(mainMB,"*button_3");
  XtVaSetValues(mainMB,XmNmenuHelpWidget,menuHelpWidget,NULL);
  for (i = 0; i < N_MAIN_MENU_ELES; i++) XmStringFree(buttons[i]);


/*
 * create the file pulldown menu pane
 */
  buttons[0] = XmStringCreateSimple("New");
  buttons[1] = XmStringCreateSimple("Open...");
  buttons[2] = XmStringCreateSimple("Save As...");
  buttons[3] = XmStringCreateSimple("Separator");
  buttons[4] = XmStringCreateSimple("Print Data...");
  buttons[5] = XmStringCreateSimple("Separator");
  buttons[6] = XmStringCreateSimple("Exit");
  keySyms[0] = 'N';
  keySyms[1] = 'O';
  keySyms[2] = 'S';
  keySyms[3] = ' ';
  keySyms[4] = 'P';
  keySyms[5] = ' ';
  keySyms[6] = 'x';
  accelerators[0] = "";
  accelerators[1] = "Ctrl<Key>O";
  accelerators[2] = "Ctrl<Key>S";
  accelerators[3] = "";
  accelerators[4] = "Ctrl<Key>P";
  accelerators[5] = "";
  accelerators[6] = "Ctrl<Key>X";
  acceleratorText[0] = XmStringCreateSimple("");
  acceleratorText[1] = XmStringCreateSimple("Ctrl+O");
  acceleratorText[2] = XmStringCreateSimple("Ctrl+S");
  acceleratorText[3] = XmStringCreateSimple("");
  acceleratorText[4] = XmStringCreateSimple("Ctrl+P");
  acceleratorText[5] = XmStringCreateSimple("");
  acceleratorText[6] = XmStringCreateSimple("Ctrl+X");
  buttonType[0] = XmPUSHBUTTON;
  buttonType[1] = XmPUSHBUTTON;
  buttonType[2] = XmPUSHBUTTON;
  buttonType[3] = XmSEPARATOR;
  buttonType[4] = XmPUSHBUTTON;
  buttonType[5] = XmSEPARATOR;
  buttonType[6] = XmPUSHBUTTON;
  n = 0;
  XtSetArg(args[n],XmNbuttonCount,N_FILE_MENU_ELES); n++;
  XtSetArg(args[n],XmNbuttons,buttons); n++;
  XtSetArg(args[n],XmNbuttonMnemonics,keySyms); n++;
  XtSetArg(args[n],XmNbuttonType,buttonType); n++;
  XtSetArg(args[n],XmNbuttonAccelerators,accelerators); n++;
  XtSetArg(args[n],XmNbuttonAcceleratorText,acceleratorText); n++;
  XtSetArg(args[n],XmNpostFromButton,FILE_BTN_POSN); n++;
  XtSetArg(args[n],XmNsimpleCallback,fileMenuSimpleCallback); n++;
  mainFilePDM = XmCreateSimplePulldownMenu(mainMB,"mainFilePDM",args,n);
  for (i = 0; i < N_FILE_MENU_ELES; i++) {
    XmStringFree(buttons[i]);
    if (acceleratorText[i] != NULL) XmStringFree(acceleratorText[i]);
  }

/*
 * create the edit pulldown menu pane
 */
  buttons[0] = XmStringCreateSimple("Cut");
  buttons[1] = XmStringCreateSimple("");
  buttons[2] = XmStringCreateSimple("Clear");
  keySyms[0] = 't';
  keySyms[1] = ' ';
  keySyms[2] = 'e';
  accelerators[0] = "Shift<Key>DeleteChar";
  accelerators[1] = "";
  accelerators[2] = "";
  acceleratorText[0] = XmStringCreateSimple("Shift+Del");
  acceleratorText[1] = XmStringCreateSimple("");
  acceleratorText[2] = XmStringCreateSimple("");
  buttonType[0] = XmPUSHBUTTON;
  buttonType[1] = XmSEPARATOR;
  buttonType[2] = XmPUSHBUTTON;
  n = 0;
  XtSetArg(args[n],XmNbuttonCount,N_EDIT_MENU_ELES); n++;
  XtSetArg(args[n],XmNbuttons,buttons); n++;
  XtSetArg(args[n],XmNbuttonMnemonics,keySyms); n++;
  XtSetArg(args[n],XmNbuttonType,buttonType); n++;
  XtSetArg(args[n],XmNbuttonAccelerators,accelerators); n++;
  XtSetArg(args[n],XmNbuttonAcceleratorText,acceleratorText); n++;
  XtSetArg(args[n],XmNpostFromButton,EDIT_BTN_POSN); n++;
  XtSetArg(args[n],XmNsimpleCallback,editMenuSimpleCallback); n++;
  mainEditPDM = XmCreateSimplePulldownMenu(mainMB,"mainEditPDM",args,n);
  for (i = 0; i < N_EDIT_MENU_ELES; i++) {
    XmStringFree(buttons[i]);
    if (acceleratorText[i] != NULL) XmStringFree(acceleratorText[i]);
  }


/*
 * create the statistics pulldown menu pane
 */
  buttons[0] = XmStringCreateSimple("Reset Selected");
  buttons[1] = XmStringCreateSimple("Reset All");
  keySyms[0] = 'R';
  keySyms[1] = 'l';
  accelerators[0] = "Ctrl<Key>R";
  accelerators[1] = "Ctrl<Key>L";
  acceleratorText[0] = XmStringCreateSimple("Ctrl+R");
  acceleratorText[1] = XmStringCreateSimple("Ctrl+L");
  buttonType[0] = XmPUSHBUTTON;
  buttonType[1] = XmPUSHBUTTON;
  n = 0;
  XtSetArg(args[n],XmNbuttonCount,N_STATISTICS_MENU_ELES); n++;
  XtSetArg(args[n],XmNbuttons,buttons); n++;
  XtSetArg(args[n],XmNbuttonMnemonics,keySyms); n++;
  XtSetArg(args[n],XmNbuttonType,buttonType); n++;
  XtSetArg(args[n],XmNbuttonAccelerators,accelerators); n++;
  XtSetArg(args[n],XmNbuttonAcceleratorText,acceleratorText); n++;
  XtSetArg(args[n],XmNpostFromButton,STATISTICS_BTN_POSN); n++;
  XtSetArg(args[n],XmNsimpleCallback,statisticsMenuSimpleCallback); n++;
  mainStatisticsPDM = XmCreateSimplePulldownMenu(mainMB,"mainStatisticsPDM",
	args,n);
  for (i = 0; i < N_STATISTICS_MENU_ELES; i++) {
    XmStringFree(buttons[i]);
    if (acceleratorText[i] != NULL) XmStringFree(acceleratorText[i]);
  }


/*
 * create the help pulldown menu pane
 */
  buttons[0] = XmStringCreateSimple("On Context");
  buttons[1] = XmStringCreateSimple("On Window...");
  buttons[2] = XmStringCreateSimple("On Keys...");
  buttons[3] = XmStringCreateSimple("Index...");
  buttons[4] = XmStringCreateSimple("On Help...");
  buttons[5] = XmStringCreateSimple("Tutorial...");
  buttons[6] = XmStringCreateSimple("On Version...");
  keySyms[0] = 'C';
  keySyms[1] = 'W';
  keySyms[2] = 'K';
  keySyms[3] = 'I';
  keySyms[4] = 'H';
  keySyms[5] = 'T';
  keySyms[6] = 'V';
  accelerators[0] = "Ctrl<Key>H"; 
  accelerators[1] = NULL;
  accelerators[2] = NULL;
  accelerators[3] = NULL;
  accelerators[4] = NULL;
  accelerators[5] = NULL;
  accelerators[6] = NULL;
  j = 0;
  acceleratorText[0] = XmStringCreateSimple("Ctrl+H"); j++;
  acceleratorText[1] = NULL;
  acceleratorText[2] = NULL;
  acceleratorText[3] = NULL;
  acceleratorText[4] = NULL;
  acceleratorText[5] = NULL;
  acceleratorText[6] = NULL;
  for (i = 0; i < 6; i++) buttonType[i] = XmPUSHBUTTON;
  buttonType[6] = XmSEPARATOR;
  buttonType[7] = XmPUSHBUTTON;

  for (i = 0; i < N_HELP_MENU_ELES; i++) buttonType[i] = XmPUSHBUTTON;
  n = 0;
  XtSetArg(args[n],XmNbuttonCount,N_HELP_MENU_ELES); n++;
  XtSetArg(args[n],XmNbuttons,buttons); n++;
  XtSetArg(args[n],XmNbuttonMnemonics,keySyms); n++;
  XtSetArg(args[n],XmNbuttonType,buttonType); n++;
  XtSetArg(args[n],XmNbuttonAccelerators,accelerators); n++;
  XtSetArg(args[n],XmNbuttonAcceleratorText,acceleratorText); n++;
  XtSetArg(args[n],XmNpostFromButton,HELP_BTN_POSN); n++;
  XtSetArg(args[n],XmNsimpleCallback,helpMenuSimpleCallback); n++;
  mainHelpPDM = XmCreateSimplePulldownMenu(mainMB,"mainHelpPDM",args,n);
  for (i = 0; i < N_HELP_MENU_ELES; i++) XmStringFree(buttons[i]);
  for (i = 0; i < j; i++) XmStringFree(acceleratorText[i]);

/* don't enable other help functions yet */
  XtSetSensitive(XtNameToWidget(mainHelpPDM,"*button_0"),False);
  XtSetSensitive(XtNameToWidget(mainHelpPDM,"*button_1"),False);
  XtSetSensitive(XtNameToWidget(mainHelpPDM,"*button_2"),False);
  XtSetSensitive(XtNameToWidget(mainHelpPDM,"*button_3"),False);
  XtSetSensitive(XtNameToWidget(mainHelpPDM,"*button_4"),False);
  XtSetSensitive(XtNameToWidget(mainHelpPDM,"*button_5"),False);


  XtAddCallback(mainMW,XmNhelpCallback,globalHelpCallback,(XtPointer)HELP_MAIN);

/*
 * manage the composites
 */
  XtManageChild(mainMB);
  XtManageChild(legendBB);
  XtManageChild(interactionForm);
  XtManageChild(mainBB);

/*
 * set the main window widgets
 */
  XmMainWindowSetAreas(mainMW,mainMB,legendBB,NULL,NULL,mainBB);
  XtVaSetValues(mainMW,XmNmessageWindow,interactionForm,NULL);
  XtManageChild(mainMW);



  XtVaGetValues(mainMW,XmNhorizontalScrollBar,&scrollBar,NULL);
  XtVaGetValues(mainMW,XmNverticalScrollBar,&verticalScrollBar,NULL);
  XtVaGetValues(mainMW,XmNclipWindow,&clipWindow,NULL);

  XtAddCallback(scrollBar,XmNdragCallback,scrollCB,NULL);
  XtAddCallback(scrollBar,XmNpageIncrementCallback,scrollCB,
				NULL);
  XtAddCallback(scrollBar,XmNpageDecrementCallback,scrollCB,
				NULL);
  XtAddCallback(scrollBar,XmNvalueChangedCallback,scrollCB,
				NULL);
  XtAddCallback(scrollBar,XmNtoTopCallback,scrollCB,
				NULL);
  XtAddCallback(scrollBar,XmNtoBottomCallback,scrollCB,
				NULL);
  XtAddCallback(scrollBar,XmNincrementCallback,scrollCB,
				NULL);
  XtAddCallback(scrollBar,XmNdecrementCallback,scrollCB,
				NULL);



/************************************************
 ****** create main-window related dialogs ******
 ************************************************/

 /*
  * note - all FILE pdm entry dialogs are sharing a single callback
  *	  fileMenuDialogCallback, with client data
  *	  of the BTN id in the simple menu...
  */


/*
 * create the file selection dialog which handles Open... and Save As...
 */
  n = 0;
  label = XmStringCreateSimple("*.dp");
  XtSetArg(args[n],XmNpattern,label); n++;
/* for some odd reason can't get PATH_MAX reliably defined between systems */
#define LOCAL_PATH_MAX	1023
  cwd = (char *)getcwd(NULL,LOCAL_PATH_MAX+1);
  cwdXmString = XmStringCreateSimple(cwd);
  XtSetArg(args[n],XmNdirectory,cwdXmString); n++;
  openFSD = XmCreateFileSelectionDialog(mainFilePDM,"openFSD",args,n);
  XtUnmanageChild(XmFileSelectionBoxGetChild(openFSD,XmDIALOG_FILTER_LABEL));
  XtUnmanageChild(XmFileSelectionBoxGetChild(openFSD,XmDIALOG_FILTER_TEXT));
  XtAddCallback(openFSD,XmNokCallback,
	fileMenuDialogCallback,(XtPointer)NULL);
  XtAddCallback(openFSD,XmNcancelCallback,
	fileMenuDialogCallback,(XtPointer)NULL);
  XmStringFree(label);
  XmStringFree(cwdXmString);
  free(cwd);

/*
 * create the Help information shell
 */
  n = 0;
  XtSetArg(args[n],XtNiconName,"Help"); n++;
  XtSetArg(args[n],XtNtitle,"dp Help System"); n++;
  XtSetArg(args[n],XtNallowShellResize,TRUE); n++;
  XtSetArg(args[n],XmNkeyboardFocusPolicy,XmPOINTER); n++;
/* map window manager menu Close function to application close... */
  helpS = XtCreatePopupShell("helpS",topLevelShellWidgetClass,
		mainShell,args,n);

  n = 0;
  XtSetArg(args[n],XmNdialogType,XmDIALOG_INFORMATION); n++;
  helpMessageBox = XmCreateMessageBox(helpS,"helpMessageBox",
			args,n);
  XtAddCallback(helpMessageBox,XmNcancelCallback,
	helpDialogCallback, (XtPointer)NULL);
  XtAddCallback(helpMessageBox,XmNokCallback,
	helpDialogCallback,(XtPointer)NULL);

  XtManageChild(helpMessageBox);


}
