/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/*
 *	Mark Anderson, Argonne National Laboratory:
 *		U.S. DOE, University of Chicago
 */

#ifndef __DP_H__
#define __DP_H__

#ifdef ALLOCATE_STORAGE
#define EXTERN
#else
#define EXTERN extern
#endif

#ifndef MAX
#  define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif
#ifndef MIN
#  define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

/* STANDARDS CONFORMANCE: AES, XPG2, XPG3, XPG4, POSIX.1, POSIX.2 */
/*
#include <sys/types.h>
#include <unistd.h>
*/
#include <limits.h>
#include <float.h> 	/* XPG4 limits.h doesn't include float.h */
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>


/*
 * X/Xt/Xm includes, globals
 */

#include <X11/IntrinsicP.h>
#include <X11/cursorfont.h>

#include <Xm/Xm.h>
#include <Xm/AtomMgr.h>
#include <Xm/DragDrop.h>
#include <Xm/MwmUtil.h>

#include <Xm/BulletinB.h>
#include <Xm/TextF.h>
#include <Xm/Label.h>
#include <Xm/FileSB.h>
#include <Xm/Frame.h>
#include <Xm/List.h>
#include <Xm/MessageB.h>
#include <Xm/MainW.h>
#include <Xm/ScrolledW.h>
#include <Xm/Separator.h>
#include <Xm/DrawingA.h>
#include <Xm/DrawnB.h>
#include <Xm/Form.h>

/*
 * DP includes
 */
#include "epicsVersion.h"
#include "dpVersion.h"

/*
 * CA includes
 */
#include "cadef.h"



/* 
 * define Xt application name/class
 */
#define NAME	"dp"
#define CLASS	"Dp"


/*
 * define the help layers
 */
#define HELP_MAIN 0


/*
 * CA related defines
 */
#define CA_PEND_EVENT_TIME	0.01



/*
 * data buffer structure defines
 */
#define CHANNEL_NAME_SIZE	64
#define OPER_RANGE_STRING_SIZE	80
#define FIELD_STRING_SIZE	80

/*
 * miscellaneous defines
 */
#define NUM_DIFFERENCES		50
#define TIME_FOR_DIFFERENCES	5.0

#define MARGIN			0	/* margins in display page layout   */
#define MOVE_MARGIN		0	/* margins used in relocating DAs   */
#define MW_MARGIN		30	/* main window margin for resizes   */

#define NOMINAL_NCOLUMNS	12	/* nominal number of columns for Ls */

#define STATUS_DIAM		7	/* diameter of status LED            */
					/* colors for status indicators      */
#define BAD_STATUS_COLOR_STRING		"red"
#define GOOD_STATUS_COLOR_STRING	"green"
#define BORDER_COLOR_STRING		"black"


typedef struct {
  double diff;		/* absolute value of difference            */
  Boolean negative;	/* True if diff is supposed to be negative */
  time_t time;  	/* # seconds since epoch...                */
} DiffTime;


/*
 * entries can be CompositeDevice (really only name, but which has >= 1 atomic
 *	devices as components)
 * or AtomicDevice (an entity which can be resolved into constituent PV's)
 * or ProcessVariable (highest level of granularity)
 */
typedef enum {
  CompositeDevice, AtomicDevice, ProcessVariable
} EntryType;

/*
 * device/PV status can be none-yet, bad, good
 */
 typedef enum {
   NoStatus, BadStatus, GoodStatus
} StatusType;


typedef struct _DpStruct {

  double setPoint;	/* acts as reference point for PV statistics too */
  double readback;
  double average;
  double rmsDiff;
  double maxDiff5;
  double maxDiff;
  double maxDiffAbs;
  double lopr;
  double hopr;
  double   averageDiffSquared;
  DiffTime diffArray[NUM_DIFFERENCES];
  int      diffArrayIndex;
  EntryType entryType;
  unsigned int count;
  StatusType status;

  Widget DA;
  Widget channelL;
  Widget setPointTF;
  Widget readbackL;
  Widget averageL;
  Widget rmsDiffL;
  Widget maxDiff5L;
  Widget maxDiffL;
  Widget unitsL;
  Widget operRangeL;

  char setPointString[FIELD_STRING_SIZE];
  char readbackString[FIELD_STRING_SIZE];
  char averageString[FIELD_STRING_SIZE];
  char rmsDiffString[FIELD_STRING_SIZE];
  char maxDiff5String[FIELD_STRING_SIZE];
  char maxDiffString[FIELD_STRING_SIZE];
  char unitsString[FIELD_STRING_SIZE];
  char operRangeString[FIELD_STRING_SIZE];

  chid   setPointChid;
  chid   readbackChid;
  chid   statusChid;
  evid   setPointEvid;
  evid   readbackEvid;
  evid   statusEvid;

  struct _DpStruct *next;
  struct _DpStruct *prev;

  short   setPointPrecision;
  short   readbackPrecision;
  Boolean memberOfComposite;
  Boolean setPointPreviouslyConnected;
  Boolean readbackPreviouslyConnected;
  Boolean statusPreviouslyConnected;
  Boolean modified;
  Boolean foundSetPoint;
  Boolean setPointChanged, readbackChanged, maxDiffChanged;
  char   units[12];		/* graphics info units from db_access.h */
  char   channel[CHANNEL_NAME_SIZE];
  char   operRangeUnitsString[OPER_RANGE_STRING_SIZE];

} DpStruct;



/*
 * global X/Xt/Motif parameters
  */
EXTERN Display *display;
EXTERN XtAppContext appContext;
EXTERN Window rootWindow;
EXTERN int screenNum;
EXTERN Atom WM_DELETE_WINDOW;
EXTERN Atom WM_TAKE_FOCUS;
EXTERN Atom COMPOUND_TEXT;
EXTERN Atom DP_DESTROY;
EXTERN XtTranslations parsedTranslations;
EXTERN Widget mainShell, helpS;
EXTERN Widget mainMW, mainMB, mainBB, mainBB1, trashCan, legendSW,
	legendBB, helpMessageBox, scrollBar, printDialog;
EXTERN Widget clipWindow,verticalScrollBar;
EXTERN Window clientMessageWindow;
EXTERN Pixmap trashCanPixmap;
EXTERN Cursor helpCursor, watchCursor;
EXTERN DpStruct *listHead, *listTail;
EXTERN int numChannels;
EXTERN GC goodStatusGC, badStatusGC, borderGC;
EXTERN int globalStatusX;

EXTERN XtIntervalId intervalId;

/* NULL terminated array of (DpStruct *)s */
EXTERN DpStruct **currentlySelectedDpStructs;

EXTERN short operRangeMaxColumns;

/*
 * widgets for legend area
 */
EXTERN Widget device, setPoint, rdbk, average, rms, diff5, diff, units,
	operRange;

EXTERN Position deviceX, setPointX, rdbkX, averageX, rmsX, diff5X, diffX,
	unitsX, operRangeX;

/*
 * function declarations
 */
void selectRow(Widget w, XtPointer client_data, XEvent *event);
extern void unselectRows(void);
extern void unselectRow(DpStruct *, DpStruct *);
extern Boolean isCurrentlySelected(DpStruct *);
extern int getNumberSelected(void);
extern void renderStatus(DpStruct *);
extern unsigned long getPixelFromString(Display *, int, char *);
void globalHelpCallback(Widget w, XtPointer client_data, XtPointer call_data);

extern void StartDrag(Widget w, XEvent *event);

/* define layout of data */
#define DEVICE			0
#define SET_POINT		1
#define RDBK			2
#define AVERAGE			3
#define RMS			4
#define DIFF5			5
#define DIFF			6
#define UNITS			7
#define OPER_RANGE		8
#define MIN_LEGEND_ENTRY	DEVICE
#define MAX_LEGEND_ENTRY	OPER_RANGE
#define N_LEGEND_ENTRIES	MAX_LEGEND_ENTRY+1

#ifndef ALLOCATE_STORAGE
  extern char *legendString[];
#else
  char *legendString[] = {"Device, Status","Set (Ref)","Read","Aver Rdbk",
				"RMS Dif", "Max Dif/5s", "Max Dif",
				"Units", "Oper Range",};
#endif




#ifndef ALLOCATE_STORAGE
  extern char *dragTranslations;
  extern XtActionsRec *dragActions;
#else
  char dragTranslations[] = "#override <Btn2Down>:StartDrag()";
  XtActionsRec dragActions[] = {{"StartDrag",(XtActionProc)StartDrag}};
#endif


#endif  /* __DP_H__ */
